/*******************************************************************************
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package com.ibm.zrh.eclipse.nls;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;

import org.eclipse.core.resources.IFile;

import org.eclipse.core.filebuffers.FileBuffers;
import org.eclipse.core.filebuffers.ITextFileBufferManager;

import org.eclipse.text.edits.ReplaceEdit;
import org.eclipse.text.edits.TextEdit;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.Position;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.Comment;
import org.eclipse.jdt.core.dom.Expression;
import org.eclipse.jdt.core.dom.FieldAccess;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.LineComment;
import org.eclipse.jdt.core.dom.MethodInvocation;
import org.eclipse.jdt.core.dom.Name;
import org.eclipse.jdt.core.dom.StringLiteral;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.SearchPattern;

import org.eclipse.jdt.internal.corext.dom.Bindings;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringScopeFactory;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringSearchEngine;
import org.eclipse.jdt.internal.corext.refactoring.changes.CompilationUnitChange;
import org.eclipse.jdt.internal.corext.refactoring.changes.TextChangeCompatibility;
import org.eclipse.jdt.internal.corext.refactoring.nls.NLSElement;
import org.eclipse.jdt.internal.corext.refactoring.structure.CompilationUnitRewrite;

import org.eclipse.jdt.internal.ui.JavaPlugin;

import org.eclipse.ltk.core.refactoring.Change;
import org.eclipse.ltk.core.refactoring.CompositeChange;
import org.eclipse.ltk.core.refactoring.Refactoring;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;

public class MessageBundleRefactoring extends Refactoring {

	IType fAccessorClass;
	IFile fPropertiesFile;
	CompositeChange fChange;
	ITypeBinding fAccessorTypeBinding;

	public MessageBundleRefactoring(IType accessorClass, IFile propertiesFile) {
		super();
		fAccessorClass = accessorClass;
		fPropertiesFile = propertiesFile;
	}

	public String getName() {
		return "Message Bundle Refactoring";
	}

	public RefactoringStatus checkInitialConditions(IProgressMonitor monitor) throws CoreException, OperationCanceledException {
		return new RefactoringStatus();
	}

	public RefactoringStatus checkFinalConditions(IProgressMonitor monitor) throws CoreException, OperationCanceledException {
		RefactoringStatus result = new RefactoringStatus();
		fAccessorTypeBinding = computeAccessorClassBinding(new SubProgressMonitor(monitor, 1));
		if (fAccessorTypeBinding == null) {
			result.addFatalError("Couldn't resolve accessor class");
			return result;
		}
		fChange = new CompositeChange("Accessor Class Changes");
		ICompilationUnit[] affectedUnits = RefactoringSearchEngine.findAffectedCompilationUnits(SearchPattern.createPattern(fAccessorClass, IJavaSearchConstants.REFERENCES), RefactoringScopeFactory.create(fAccessorClass), new SubProgressMonitor(monitor, 5), result);
		monitor.beginTask("", affectedUnits.length + 1);
		for (int i = 0; i < affectedUnits.length; i++) {
			ICompilationUnit unit = affectedUnits[i];
			if (unit.equals(fAccessorClass.getCompilationUnit()))
				continue;
			processCompilationUnit(result, unit, new SubProgressMonitor(monitor, 1));
		}
		processPropertiesFile(result, new SubProgressMonitor(monitor, 1));
		return result;
	}

	private void processPropertiesFile(RefactoringStatus result, IProgressMonitor monitor) throws CoreException {
		// TODO need to roll the changes to the properties file into a text edit that we can hook into
		// the "Undo -> Refactoring" framework
		try {
			ITextFileBufferManager manager = FileBuffers.getTextFileBufferManager();
			try {
				manager.connect(fPropertiesFile.getFullPath(), null);
				manager.connect(fAccessorClass.getCompilationUnit().getCorrespondingResource().getFullPath(), null);
				fChange.addAll(new PropertyFileConverter().convertFile(fAccessorClass, fPropertiesFile));
			} finally {
				manager.disconnect(fPropertiesFile.getFullPath(), null);
				manager.disconnect(fAccessorClass.getCompilationUnit().getCorrespondingResource().getFullPath(), null);
			}
		} catch (IOException e) {
			throw new CoreException(new Status(IStatus.ERROR, NLSPlugin.PI_NLS, IStatus.ERROR, e.getMessage(), e));
		}
	}

	private void processCompilationUnit(RefactoringStatus result, ICompilationUnit unit, IProgressMonitor monitor) throws CoreException {
		monitor.beginTask("", 2);
		CompilationUnitRewrite rewrite= new CompilationUnitRewrite(unit);
		List edits= new ArrayList();
		processAST(edits, result, rewrite, new SubProgressMonitor(monitor, 1));
		final CompilationUnitChange change= rewrite.createChange();
		if (!edits.isEmpty()) {
			for (final Iterator iterator= edits.iterator(); iterator.hasNext();) {
				TextEdit edit= (TextEdit) iterator.next();
				TextChangeCompatibility.addTextEdit(change, "", edit);
			}
		}
		fChange.add(change);
		monitor.done();
	}

	private void processAST(final List edits, final RefactoringStatus result, final CompilationUnitRewrite rewriter, SubProgressMonitor monitor) throws JavaModelException {
		// Quick and dirty...
		final List comments= rewriter.getRoot().getCommentList();
		final IDocument document= new Document(rewriter.getCu().getBuffer().getContents());
		ASTVisitor visitor= new ASTVisitor() {

			public void endVisit(MethodInvocation node) {
				Name messageBundleName= getMessageBundleReceiver(node);
				if (messageBundleName == null)
					return;
				IMethodBinding method= node.resolveMethodBinding();
				ITypeBinding returnType= method.getReturnType();
				if (!returnType.getQualifiedName().equals("java.lang.String"))
					return;
				ITypeBinding[] params= method.getParameterTypes();
				if (params.length < 1 && params.length > 2)
					return;
				for (int index= 0; index < params.length; index++) {
					final String qualifiedName= params[index].getQualifiedName();
					if (!(qualifiedName.startsWith("java.lang.Object") || qualifiedName.startsWith("java.lang.String")))
						return;
				}
				List args= node.arguments();

				// MessageBundle#getString(String)
				if (args.size() == 1 && method.getName().equals("getString")) {
					Object obj= args.get(0);
					if (!(obj instanceof StringLiteral))
						return;
					StringLiteral string= (StringLiteral) obj;
					String key= PropertyFileConverter.convertToJavaIdentifier(string.getLiteralValue());

					FieldAccess fieldAccess= rewriter.getRoot().getAST().newFieldAccess();
					fieldAccess.setExpression((Expression) rewriter.getASTRewrite().createCopyTarget(messageBundleName));
					fieldAccess.setName(rewriter.getRoot().getAST().newSimpleName(key));

					rewriter.getASTRewrite().replace(node, fieldAccess, null);

					try {
						removeComments(edits, comments, node, document);
					} catch (BadLocationException exception) {
						JavaPlugin.log(exception);
					}

					// MessageBundle#getFormattedString(String, Object)
					// MessageBundle#getFormattedString(String, String)
					// MessageBundle#getFormattedString(String, Object[])
					// MessageBundle#getFormattedString(String, String[])
				} else if (args.size() == 2 && method.getName().equals("getFormattedString")) {
					Object obj= args.get(0);
					if (!(obj instanceof StringLiteral))
						return;
					Expression expression= (Expression) args.get(1);
					ITypeBinding argBinding= expression.resolveTypeBinding();
					if (argBinding == null)
						return;

					StringLiteral string= (StringLiteral) obj;
					String key= PropertyFileConverter.convertToJavaIdentifier(string.getLiteralValue());

					FieldAccess fieldAccess= rewriter.getAST().newFieldAccess();
					fieldAccess.setExpression((Expression) rewriter.getASTRewrite().createCopyTarget(messageBundleName));
					fieldAccess.setName(rewriter.getAST().newSimpleName(key));

					rewriter.getASTRewrite().replace((ASTNode) obj, fieldAccess, null);
					rewriter.getASTRewrite().replace(node.getName(), rewriter.getAST().newSimpleName("bind"), null); //$NON-NLS-1$
					rewriter.getASTRewrite().replace(messageBundleName, rewriter.getImportRewrite().addImportFromSignature(Signature.createTypeSignature("org.eclipse.osgi.util.NLS", true), rewriter.getAST()), null);

					try {
						removeComments(edits, comments, node, document);
					} catch (BadLocationException exception) {
						JavaPlugin.log(exception);
					}

				} else
					result.merge(RefactoringStatus.createFatalErrorStatus("The method invocation '" + node.toString() + "' could not be converted"));

				return;
			}

			private Name getMessageBundleReceiver(MethodInvocation node) {
				Expression expression= node.getExpression();
				if (expression == null)
					return null;
				if (expression instanceof Name && Bindings.equals(fAccessorTypeBinding, ((Name) expression).resolveBinding())) {
					return (Name) expression;
				}
				return null;
			}
		};
		rewriter.getRoot().accept(visitor);
	}
	
	private void removeComments(List edits, List comments, ASTNode node, IDocument document) throws BadLocationException {
		final IRegion region= document.getLineInformationOfOffset(node.getStartPosition());
		final Position position= new Position(region.getOffset(), region.getLength());
		for (final Iterator iterator= comments.iterator(); iterator.hasNext();) {
			final Comment comment= (Comment) iterator.next();
			if (comment instanceof LineComment) {
				Position range= new Position(comment.getStartPosition(), comment.getLength());
				String content= document.get(comment.getStartPosition(), comment.getLength());
				if (content.startsWith(NLSElement.TAG_PREFIX) && position.includes(range.offset))
					edits.add(new ReplaceEdit(comment.getStartPosition(), comment.getLength(), ""));
			}
		}
	}
	
	public Change createChange(IProgressMonitor monitor) throws CoreException, OperationCanceledException {
		return fChange;
	}

	private ITypeBinding computeAccessorClassBinding(IProgressMonitor monitor) {
		ASTParser parser = ASTParser.newParser(AST.JLS3);
		parser.setProject(fAccessorClass.getJavaProject());
		return (ITypeBinding) parser.createBindings(new IJavaElement[] {fAccessorClass}, monitor)[0];
	}
}
